/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

/******************************* C H E C K *******************************/

static do_check_scalar_numeric_expr (expr, c)
Node *expr;
char c;
{
	int type = expr->com.typ->etype->type;
	int rank = RANK_FROM_SHAPE (expr->com.typ->shape);

	if (type != TYP_INTEGER && type != TYP_REAL) {
		sprintf (serr, "expression %c in DO control must be of type INTEGER or REAL, is %s"
		, c, etype_name (expr->com.typ->etype));
		fatal (serr);
	}

	if (rank) {
		sprintf (serr, "expression %c in DO control must be scalar, is of rank %d"
		, c, rank);
		fatal (serr);
	}
}

Node *check_DoStmt (root)
Node *root;
{
	struct S_DoStmt *dos = &root->DoStmt;
	Node *doc;
	struct S_DoConstruct *sdoc;

	doc = make (DO_CONSTRUCT);
	sdoc = &doc->DoConstruct;
	block_push (doc);
	BLOCK_ADD_STMTS_TO (sdoc->DoBlock);
	sdoc->DoStmt = root;

	sdoc->label = se_current_statement->Statement.IntegerConstant;
	sdoc->file = input_filename;
	sdoc->line = zzlineno;
	sdoc->label_do = dos->OptLabel;

/* checks DO with LoopControl */

	if (dos->LoopControl != NN) {
		Node *Nname;

		if (dos->LoopControl->LoopControl.ImpliedDoControl != NN) {
			struct S_ImpliedDoControl *impdo;
			Node *var;
			int type, rank;
	
			impdo = &dos->LoopControl->LoopControl.ImpliedDoControl->ImpliedDoControl;
			var = impdo->Name;
			type = var->com.typ->etype->type;
			rank = RANK_FROM_SHAPE (var->com.typ->shape);
	
			if (type != TYP_INTEGER && type != TYP_REAL) {
				sprintf (serr, "DO variable must be of type INTEGER or REAL, is %s"
				, etype_name (var->com.typ->etype));
				fatal (serr);
			}
	
			if (rank) {
				sprintf (serr, "DO variable must be scalar, is of rank %d", rank);
				fatal (serr);
			}
	
			do_check_scalar_numeric_expr (impdo->Expr, '1');
			do_check_scalar_numeric_expr (impdo->Expr1, '2');
			if (impdo->Expr2 != NN)
				do_check_scalar_numeric_expr (impdo->Expr2, '3');
		}
		else if ((Nname = dos->LoopControl->LoopControl.Name) != NN) { /*WHILE or TIMES*/
			Node *expr = dos->LoopControl->LoopControl.Expr;
			Etype *etype = expr->com.typ->etype;
			int rank = RANK_FROM_SHAPE (expr->com.typ->shape);

			if (!U_strcmp_nocase (Nname->com.str, "WHILE")) {
				if (etype->type != TYP_LOGICAL) {
					sprintf (serr, "expression in DO WHILE must be of type LOGICAL, is %s"
					, etype_name (etype));
					fatal (serr);
				}
	
				if (rank) {
					sprintf (serr, "expression in DO WHILE must be scalar, is of rank %d"
					, rank);
					fatal (serr);
				}
			}
			else {  /* is TIMES */
				stand ("DO TIMES construct");

				if (etype->type != TYP_INTEGER) {
					sprintf (serr, "expression in DO TIMES must be of type INTEGER, is %s"
					, etype_name (etype));
					fatal (serr);
				}
	
				if (rank) {
					sprintf (serr, "expression in DO TIMES must be scalar, is of rank %d"
					, rank);
					fatal (serr);
				}
			}
		}
	}

	IGNORE_RETURN;
}

Node *check_EndDoStmt (root)
Node *root;
{
	Node *doc, *enddo_label;
	struct S_DoConstruct *sdoc;
	char *do_name, *enddo_name;

	if ((doc = block_match ("END DO", DO_CONSTRUCT)) == NN)
		IGNORE_RETURN;

	doc = block_pop ();
	sdoc = &doc->DoConstruct;
	sdoc->EndDoStmt = root;
	sdoc->label_enddo = CURRENT_LABEL;

/* checks construct-name stuff */

	do_name = (sdoc->DoStmt->DoStmt.Name != NN) ? sdoc->DoStmt->DoStmt.Name->com.str : CN;
	enddo_name = (root->EndDoStmt.Name != NN) ? root->EndDoStmt.Name->com.str : CN;

	if (do_name != CN && enddo_name != CN) {
		if (U_strcmp_nocase (do_name, enddo_name))
			fatal ("construct-name in END DO doesn't match that of the DO");
	}
	else if (do_name != CN && enddo_name == CN) {
		fatal ("missing construct-name in END DO statement");
	}
	else if (do_name == CN && enddo_name != CN) {
		fatal ("this END DO should not have a construct-name");
	}

/* checks label stuff */

	enddo_label = se_current_statement->Statement.IntegerConstant;

	if (sdoc->label_do != NN)
		if (enddo_label == NN
		||  strcmp (enddo_label->com.str, sdoc->label_do->com.str)) {
			sprintf (serr, "END DO statement must have label %s"
			, sdoc->label_do->com.str);
			fatal (serr);
		}

/* clears */

	doc = trans_DoConstruct (doc);  /*!!! eventually done in trans_pass */

	block_add_stmt_with_label (doc, sdoc->label, sdoc->file, sdoc->line);
	sdoc->trans = TRN;

	IGNORE_RETURN;
}

Node *do_check_Statement (label, stmt)
Node *label, *stmt;
{
	Node *doc, *Nname, *label_cycle, *label_exit;
	struct S_DoConstruct *sdoc;
	int matched = 0, stmt_type;

	if (label == NN)
		return stmt;

/* Looks to see if this STMT matches any DO loops in the stack */

	if ((doc = block_match_continue (label)) == NN)
		return stmt;  /* if it doesn't match just returns normal statement */

	if ((Nname = doc->DoConstruct.DoStmt->DoStmt.Name) != NN) {
		sprintf (serr
		, "DO termination must be an END DO statement with construct-name '%s'"
		, Nname->com.str);
		fatal (serr);
		return stmt;
	}
		
	while ((doc = block_match_continue (label)) != NN) {
		doc = block_pop ();
		sdoc = &doc->DoConstruct;

		if (!matched) {
			sdoc->EndDoStmt = stmt;
			sdoc->label_enddo = label;
		}

		matched++;

		doc = trans_DoConstruct (doc);  /*!!! eventually done in trans_pass */

		block_add_stmt_with_label (doc, sdoc->label, sdoc->file, sdoc->line);
		sdoc->trans = TRN;
	}

	stmt_type = stmt->com.type;

	if (!order_is_action_stmt (stmt_type)
	||  stmt_type == GOTO_STMT
	||  stmt_type == RETURN_STMT
	||  stmt_type == STOP_STMT
	||  stmt_type == EXIT_STMT
	||  stmt_type == CYCLE_STMT
	||  stmt_type == END_FUNCTION_STMT
	||  stmt_type == END_SUBROUTINE_STMT
	||  stmt_type == END_PROGRAM_STMT
	||  stmt_type == ARITHMETIC_IF_STMT
	||  stmt_type == ASSIGNED_GOTO_STMT
	) {
		if (matched == 1) {
			sprintf (serr, "DO termination action-statement cannot be a %s"
			, node_Symbol (stmt_type));
		}
		else {
			sprintf (serr, "DO termination shared-statement cannot be a %s"
			, node_Symbol (stmt_type));
		}

		fatal (serr);
	}

	trans_pass (stmt);  /* necessary */

	IGNORE_RETURN;
}

Node *check_CycleStmt (root)
Node *root;
{
	Node *doc;
	struct S_DoConstruct *sdoc;

	if (root->CycleStmt.Name == NN) {
		if ((doc = block_match_exit (CN, CYCLE)) == NN)
			return root;

		sdoc = &doc->DoConstruct;
		sdoc->cycle_list = list_add_end (sdoc->cycle_list, root, SUBSCRIPT_LIST);
	}
	else {
		if ((doc = block_match_exit (root->CycleStmt.Name->com.str, CYCLE)) == NN)
			return root;

		sdoc = &doc->DoConstruct;
		sdoc->cycle_list = list_add_end (sdoc->cycle_list, root, SUBSCRIPT_LIST);
	}

	return root;
}

Node *check_ExitStmt (root)
Node *root;
{
	Node *doc;
	struct S_DoConstruct *sdoc;

	if (root->ExitStmt.Name == NN) {
		if ((doc = block_match_exit (CN, EXIT)) == NN)
			return root;

		sdoc = &doc->DoConstruct;
		sdoc->exit_list = list_add_end (sdoc->exit_list, root, SUBSCRIPT_LIST);
	}
	else {
		if ((doc = block_match_exit (root->ExitStmt.Name->com.str, EXIT)) == NN)
			return root;

		sdoc = &doc->DoConstruct;
		sdoc->exit_list = list_add_end (sdoc->exit_list, root, SUBSCRIPT_LIST);
	}

	return root;
}

/******************************* T R A N S *******************************/

Node *trans_DoConstruct (node)
Node *node;
{
	struct S_DoConstruct *sdoc = &node->DoConstruct;
	Node *ele, *temp, *Nname, *dummy;

	sdoc->label_cycle = label_make ();
	sdoc->label_exit = label_make ();

	for (LIST_LOOP (ele, sdoc->exit_list, temp))
		ele->ExitStmt.label = sdoc->label_exit;

	for (LIST_LOOP (ele, sdoc->cycle_list, temp))
		ele->CycleStmt.label = sdoc->label_cycle;

	if (sdoc->DoStmt->DoStmt.LoopControl != NN
	&& (Nname = sdoc->DoStmt->DoStmt.LoopControl->LoopControl.Name) != NN
	&& !U_strcmp_nocase (Nname->com.str, "TIMES"))
		sdoc->do_times_var = temp_make (Typ_INTEGER_0_shape_0, 1, &dummy);

	return node;
}
	
/******************************* W R I T E *******************************/

#define DO_LOOP		1
#define DO_WHILE	2
#define DO_PLAIN	3
#define DO_TIMES	4

write_DoConstruct (node)
Node *node;
{
	struct S_DoConstruct *sdoc = &node->DoConstruct;
	struct S_DoStmt *sdos = &sdoc->DoStmt->DoStmt;
	Node *Nname;

	int has_exit, has_cycle, extra_do, extra_enddo, do_type;

/* calculates cases */

	has_exit = (sdoc->exit_list != NN) ? 1 : 0;
	has_cycle = (sdoc->cycle_list != NN) ? 1 : 0;
	extra_do = trans_tree_has_stmts (sdoc->DoStmt);
	extra_enddo = trans_tree_has_stmts (sdoc->EndDoStmt);

	if (sdos->LoopControl == NN)
		do_type = DO_PLAIN;
	else if ((Nname = sdos->LoopControl->LoopControl.Name) == NN)
		do_type = DO_LOOP;
	else if (!U_strcmp_nocase (Nname->com.str, "WHILE"))
		do_type = DO_WHILE;
	else
		do_type = DO_TIMES;

/* writes extra statements for DoStmt */

	sta_write_sta_bef (sdoc->DoStmt);

/* writes DO head */

	switch (do_type) {

	case DO_LOOP:
		oTab ();
		sdos->OptLabel = sdoc->label_cycle;
		onod (sdoc->DoStmt);
		oDon ();
		break;

	case DO_WHILE:
		ostart ();
		olab (sdoc->label_cycle);
		otab ();
		oIF ();
		oLP ();
		oNOT_OP ();
		oLP ();
		onod (sdoc->DoStmt->DoStmt.LoopControl->LoopControl.Expr);
		oRP ();
		oRP ();
		oGO_TO ();
		olab (sdoc->label_exit);
		oDon ();
		break;

	case DO_PLAIN:
		o_continue (sdoc->label_cycle);
		break;

	case DO_TIMES:
		oTab ();
		oDO ();
		olab (sdoc->label_cycle);
		onod (sdoc->do_times_var);
		oEQ ();
		oone ();
		oCOMMA ();
		onod (sdoc->DoStmt->DoStmt.LoopControl->LoopControl.Expr);
		oDon ();
		break;
	}

/* writes DO body */

	write_pass (sdoc->DoBlock);

/* writes DO tail */

	if (sdoc->EndDoStmt != NN  /* will happen if outer-shared-do-construct */
	&& sdoc->EndDoStmt->com.type != END_DO_STMT)
		write_pass (make_Statement (sdoc->label_enddo, sdoc->EndDoStmt));
	else if (sdoc->label_enddo != NN)
		write_pass (make_Statement (sdoc->label_enddo, make_ContinueStmt ()));

	switch (do_type) {

	case DO_LOOP:
	case DO_TIMES:
		o_continue (sdoc->label_cycle);
		break;

	case DO_WHILE:
	case DO_PLAIN:
	o_goto (sdoc->label_cycle);
		break;
	}

	if (has_exit || do_type == DO_WHILE)
		o_continue (sdoc->label_exit);

	sta_write_sta_aft (sdoc->DoStmt);
}

write_DoStmt (node)
Node *node;
{
	oDO ();
	if (node->DoStmt.OptLabel != NN)
		olab (node->DoStmt.OptLabel);
	write_pass (node->DoStmt.LoopControl);
}

write_ExitStmt (node)
Node *node;
{
	oGO_TO ();
	olab (node->ExitStmt.label);
}

write_CycleStmt (node)
Node *node;
{
	oGO_TO ();
	olab (node->CycleStmt.label);
}
